#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os


class Settings:

    __SPECIFIED__ = False

    _root = os.path.abspath(os.path.dirname(__file__))
    _base = os.path.abspath(os.path.dirname(_root))

    # 日志文件存放的目录
    LOG_DIR = os.path.join(_base, "logs")
    # 报告文件存放的目录
    REPORT_DIR = os.path.join(_base, "reports")
    # 需要执行测试用例（python文件）的目录地址或python文件地址
    TEST_SUITE_PATH = os.path.join(_root, 'demo', 'testsuite')
    # 需要执行测试用例（excel文件）的地址
    TEST_EXCEL_PATH = os.path.join(_root, "excel_testset", "TestCases.xlsx")
    # 工程根目录路径
    PROJECT_ROOT = os.path.dirname(TEST_SUITE_PATH)

    # 测试用例失败截图函数
    CAP_FUNC = None
    # 测试用例失败截图存放目录
    CAP_DIR = os.path.join(_base, "screenshots")
    # 测试用例失败截图函数参数
    CAP_KWARGS = {}

    # 过滤日志的级别
    LOG_CONSOLE_LEVEL = "DEBUG"
    LOG_FILE_LEVEL = "DEBUG"
    LOG_REPORT_LEVEL = "DEBUG"

    # 是否开启日志的开关
    LOG_CONSOLE_SWITCH = True
    LOG_FILE_SWITCH = False
    LOG_REPORT_SWITCH = False

    # 测试用例执行次数（>=1 则失败后会重新执行）
    CASE_FAIL_RERUN = 1
    # 测试用例集形式，"CODE", "EXCEL", "API"
    TEST_SET_FORM = "CODE"
    # 执行包含如下Tags的测试用例，忽略大小写
    RUN_TAGS = ["ALL"]

    # 测试类和测试方法的前缀，忽略大小写
    TEST_PREFIX = 'test'
    # 需要跳过执行的目录和文件的前缀，忽略大小写
    SKIP_PREFIX = 'puskip'
    # 日志显示断言的参数的长度
    ASSERT_ARG_LENGTH = 50



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
from punittest import SETTINGS, RELOAD_SETTINGS
from punittest import logger, TestRunner

demo = os.path.abspath(os.path.dirname(__file__))
root = os.path.abspath(os.path.dirname(demo))
# 先修改punittest的设置然后重载
SETTINGS.PROJECT_ROOT = os.path.abspath(os.path.dirname(root))
SETTINGS.RUN_TAGS = ["ALL"]
SETTINGS.CASE_FAIL_RERUN = 1
SETTINGS.TEST_PREFIX = 'test'
SETTINGS.SKIP_PREFIX = 'skip'

# 如果是TEST_SET_FORM是CODE则需要指定TEST_SUITE_DIR
SETTINGS.TEST_SET_FORM = "CODE"
SETTINGS.TEST_SUITE_PATH = os.path.join(demo, 'testsuite', 'login')
SETTINGS.TEST_EXCEL_PATH = os.path.join(root, 'excel_testset', 'TestCases.xlsx')

SETTINGS.LOG_CONSOLE_SWITCH = True
SETTINGS.LOG_FILE_SWITCH = False
SETTINGS.LOG_REPORT_SWITCH = False
SETTINGS.LOG_CONSOLE_LEVEL = "DEBUG"
SETTINGS.LOG_FILE_LEVEL = "DEBUG"
SETTINGS.LOG_REPORT_LEVEL = "INFO"

SETTINGS.LOG_DIR = r"D:\Temp\Logs"
SETTINGS.REPORT_DIR = r"D:\Temp\Reports"

SETTINGS.CAP_FUNC = lambda _dir, name, **kwargs: logger\
    .info(r"创建截图{}\{}，参数{}".format(_dir, name, kwargs))
SETTINGS.CAP_DIR = r"D:\Temp\screenshots"
SETTINGS.CAP_KWARGS = {"arg1": "val1", "arg2": "val2"}

RELOAD_SETTINGS()

# 运行测试
runner = TestRunner("demo接口测试用例")
results = runner.run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import time
import os

from ..settings import Settings
from logging import FileHandler
from io import StringIO


class Logger:

    def __init__(self, logger_name="punittest",
                 logger_dir=Settings.LOG_DIR,
                 console_level=Settings.LOG_CONSOLE_LEVEL,
                 file_level=Settings.LOG_FILE_LEVEL,
                 report_level=Settings.LOG_REPORT_LEVEL):
        """
        :param logger_dir: 日志文件所在目录路径
        :param logger_name: logger名称
        :param console_level: 控制台输出级别
        :param file_level: 文件输出级别
        :param report_level: 报告输出级别
        """
        self.logger = logging.getLogger(logger_name)
        # 将Log放入StringIO，最后由HTMLTestRunner读取
        self.logger.log_capture = StringIO()
        self.logger.log_file_name = None
        self.logger_dir = logger_dir
        logging.root.setLevel(logging.NOTSET)
        self.formatter = logging.Formatter(
            # '[%(asctime)s][%(levelname)s][%(module)s][%(funcName)s][%(lineno)d]: %(message)s')
            '[%(asctime)s][%(levelname)s]: %(message)s')
        self.console_output_level = console_level
        self.file_output_level = file_level
        self.report_output_level = report_level
        self.backup_count = 20

    def add_handlers(self, console_switch=Settings.LOG_CONSOLE_SWITCH,
                   file_switch=Settings.LOG_FILE_SWITCH,
                   report_switch=Settings.LOG_REPORT_SWITCH):
        """
        在logger中添加日志句柄并返回，如果logger已有句柄，则直接返回
        :param console_switch: 控制台日志开关
        :param file_switch: 文件日志开关
        :param report_switch: 报告日志开关
        :return:
        """
        if console_switch:
            console_handler = logging.StreamHandler()
            console_handler.setLevel(self.console_output_level)
            console_handler.setFormatter(self.formatter)
            self.logger.addHandler(console_handler)
        if file_switch:
            now_time = time.strftime("%Y-%m-%d_%H-%M-%S")
            self.logger.log_file_name = '{0}.log'.format(now_time)
            if not os.path.exists(self.logger_dir):
                os.makedirs(self.logger_dir)
            log_file_path = os.path.join(self.logger_dir, self.logger.log_file_name)
            file_handler = FileHandler(filename=log_file_path, delay=True, encoding='utf-8')
            file_handler.setLevel(self.file_output_level)
            file_handler.setFormatter(self.formatter)
            self.logger.addHandler(file_handler)
        if report_switch:
            # 将Log放入StringIO，最后由HTMLTestRunner读取
            report_handler = logging.StreamHandler(getattr(self.logger, 'log_capture'))
            report_handler.setLevel(self.report_output_level)
            report_handler.setFormatter(self.formatter)
            self.logger.addHandler(report_handler)
        return self.logger

    def get_logger(self):
        if self.logger.handlers:
            return self.logger
        else:
            return self.add_handlers()

    def get_new_logger(self, console_switch, file_switch, report_switch):
        self.logger.handlers.clear()
        return self.add_handlers(console_switch, file_switch, report_switch)


logger = Logger().get_logger()



#! /usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import datetime
import shutil

from .settings import Settings
from .testset import ApiTestSet, CodeTestSet, ExcelTestSet
from .testcase import TestCaseHandler
from .testresult import TestResult
from .utils.logger import logger


result_data = dict()
result_data['testResult'] = []
STATUS = {
    0: 'Pass',
    1: 'Fail',
    2: 'Error',
    3: 'Skip',
}


class TestRunner:
    def __init__(self, report_title, verbosity=1, description=""):
        self.verbosity = verbosity
        self.title = report_title
        self.description = description
        self.start_time = datetime.datetime.now()
        self.stop_time = None

    def _run(self, test):
        msg = "开始测试，测试标签[{}]，用例数量总共{}个，跳过{}个，实际运行{}个"
        logger.info(msg.format(
            ', '.join(Settings.RUN_TAGS),
            TestCaseHandler.total_case_num,
            TestCaseHandler.total_case_num - TestCaseHandler.actual_case_num,
            TestCaseHandler.actual_case_num
        ))
        logger.info("")
        result = TestResult(self.verbosity)
        test(result)
        self.stop_time = datetime.datetime.now()
        self.analyze_test_result(result)
        logger.info('')
        logger.info('执行完毕: 成功{}，跳过{}，失败{}，报错{}'.format(
            result_data['testPass'], result_data['testSkip'],
            result_data['testFail'], result_data['testError']
        ))
        logger.info('测试时间: {}'.format(self.stop_time - self.start_time))
        logger.info('')
        logger.info('失败用例:')
        for tr in result_data['testResult']:
            if tr['status'] == '失败':
                logger.info('{} - {} - {}'.format(tr['moduleName'], tr['className'], tr['methodName']))
        logger.info('')
        logger.info('报错用例:')
        for tr in result_data['testResult']:
            if tr['status'] == '报错':
                logger.info('{} - {} - {}'.format(tr['moduleName'], tr['className'], tr['methodName']))
        if Settings.LOG_REPORT_SWITCH:
            report_dir = Settings.REPORT_DIR
            if not os.path.exists(report_dir):
                os.mkdir(report_dir)
            root = os.path.abspath(os.path.dirname(__file__))
            static_src_path = os.path.join(root, "static")
            static_tar_path = os.path.join(report_dir, "static")
            if not os.path.exists(static_tar_path):
                shutil.copytree(static_src_path, static_tar_path)
            from .report_template import style_2
            file_path = os.path.join(report_dir, r"{}.html".format(self.start_time.strftime("%Y-%m-%d_%H-%M-%S")))
            style_2.build_report(file_path, result_data)

    @staticmethod
    def sort_result(case_results):
        rmap = {}
        classes = []
        for n, t, o, e, run_time, start_time, end_time in case_results:
            cls = t.__class__
            if cls not in rmap:
                rmap[cls] = []
                classes.append(cls)
            rmap[cls].append((n, t, o, e, run_time, start_time, end_time))
        r = [(cls, rmap[cls]) for cls in classes]
        return r

    def analyze_test_result(self, result):
        result_data["reportName"] = self.title
        result_data["beginTime"] = str(self.start_time)
        result_data["finishTime"] = str(self.stop_time)
        result_data["totalTime"] = str(self.stop_time - self.start_time)[0: -5] + "s"

        sorted_result = self.sort_result(result.result)
        for cid, (cls, cls_results) in enumerate(sorted_result):
            pass_num = fail_num = error_num = skip_num = 0
            for case_state, *_ in cls_results:
                if case_state == 0:
                    pass_num += 1
                elif case_state == 1:
                    fail_num += 1
                elif case_state == 2:
                    error_num += 1
                else:
                    skip_num += 1

            for tid, (state_id, t, o, e, run_time, start_time, end_time) in enumerate(cls_results):
                name = t.id().split('.')[-1]
                doc = t.shortDescription() or ""
                case_data = dict()
                case_data['moduleName'] = cls.__module__
                case_data['className'] = cls.__name__
                case_data['methodName'] = name
                case_data['spendTime'] = str(run_time)[0: -5] + "s"
                case_data['startTime'] = str(start_time)
                case_data['endTime'] = str(end_time)
                case_data['description'] = doc
                case_data['log'] = o + e
                if STATUS[state_id] == "Pass":
                    case_data['status'] = "成功"
                if STATUS[state_id] == "Fail":
                    case_data['status'] = "失败"
                if STATUS[state_id] == "Error":
                    case_data['status'] = "报错"
                if STATUS[state_id] == "Skip":
                    case_data['status'] = "跳过"
                result_data['testResult'].append(case_data)

        result_data["testPass"] = result.success_count
        result_data["testAll"] = result.success_count + result.failure_count + result.error_count + result.skip_count
        result_data["testFail"] = result.failure_count
        result_data["testSkip"] = result.skip_count
        result_data["testError"] = result.error_count

    def run(self, api_data=None):
        if Settings.TEST_SET_FORM == "CODE":
            test_set = CodeTestSet(Settings.TEST_PREFIX, Settings.SKIP_PREFIX)
        elif Settings.TEST_SET_FORM == "EXCEL":
            test_set = ExcelTestSet(Settings.TEST_PREFIX, Settings.SKIP_PREFIX)
        elif Settings.TEST_SET_FORM == "API":
            test_set = ApiTestSet(api_data, Settings.TEST_PREFIX, Settings.SKIP_PREFIX)
        else:
            test_set = CodeTestSet(Settings.TEST_PREFIX, Settings.SKIP_PREFIX)
        suite = test_set.convert_test_suite()
        self._run(suite)
        return result_data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
import re
import os
import platform

from .settings import Settings
from .utils.parse_excel import ExcelParser
from .utils.logger import logger
from importlib import import_module


class TestSet:
    def __init__(self, test_prefix, skip_prefix):
        self.test_set = []
        self.case_prefix = test_prefix.lower()
        self.skip_prefix = skip_prefix.lower()
        self.test_suite = unittest.TestSuite()

    @staticmethod
    def convert_to_pydata(string):
        """将测试用例中的字符串转换成python数据结构"""
        if string is None:
            return None
        elif isinstance(string, bool):
            return [string]
        else:
            param_list = string.split('\n')
            try:
                param_list = [eval(x.strip()) for x in param_list]
            except NameError:
                param_list = [x.strip() for x in param_list]
            return param_list

    @staticmethod
    def convert_params(params):
        """将测试用例中的ParamsData由字符串转换成python数据结构"""
        return TestSet.convert_to_pydata(params)

    @staticmethod
    def convert_asserts(asserts):
        """将测试用例中的AssertResult由字符串转换成python数据结构"""
        return TestSet.convert_to_pydata(asserts)

    @staticmethod
    def convert_tags(tags):
        """将测试用例中的Tag解析成列表"""
        # 将字符串移除换行符，并利用逗号分隔
        tag_list = ['all']
        if tags is not None:
            tags = [t.strip() for t in tags.replace('\n', '').split(',')]
            if len(tags) > 0:
                # 全部替换为小写
                tags = [tag.lower() for tag in tags]
                tag_list.extend(tags)
        return tag_list

    def filter_test_case(self, **titles):
        """
        根据表格中的title过滤测试用例
        :param titles: 关键字参数，每个参数名对应表格中的title
        :return: 测试用例组成的列表
        """
        if titles:
            for title, value in titles.items():
                self.test_set = list(filter(lambda d: title in d.keys(), self.test_set))
                self.test_set = list(filter(lambda d: d[title] == value, self.test_set))
        return self.test_set

    @staticmethod
    def get_import_list_by_path(path):
        """
        根据文件/目录的绝对路径转化为import_list
        :param path:
        :return: 可被import_module识别的字符串
        """
        root, file = os.path.split(path)
        if os.path.splitext(file)[-1] == '.py' and file != '__init__.py':
            root_dir_name = os.path.split(Settings.PROJECT_ROOT)[-1]
            if platform.system() == 'Windows':
                path_list = root.split('\\')
            else:
                path_list = root.split('/')
            path_list = path_list[path_list.index(root_dir_name) + 1:]
            path_list.append(os.path.splitext(file)[0])
            import_list = '.'.join(path_list)
            return import_list
        else:
            return None

    def load_test_set(self, *args, **kwargs):
        raise NotImplementedError

    def convert_test_suite(self, *args, **kwargs):
        raise NotImplementedError


class CodeTestSet(TestSet):
    """从本地代码处获取测试集"""
    def __init__(self, test_prefix, skip_prefix):
        super(CodeTestSet, self).__init__(test_prefix, skip_prefix)
        self.test_set = None

    def _load_test_case(self, package):
        """
        加载一个python package中的所有测试用例
        :param package: package
        :return:
        """
        for cls_name in package.__dict__:
            if cls_name.lower().startswith(self.case_prefix):
                cls = getattr(package, cls_name)
                for name, func in list(cls.__dict__.items()):
                    _name = re.sub(r'_#[\d]*', '', name)
                    _name = re.sub(self.case_prefix + r'_\d{5}', self.case_prefix, _name)
                    if _name.startswith(self.case_prefix):
                        case = cls(name)
                        self.test_suite.addTest(case)

    def load_test_set(self, package):
        return self._load_test_case(package)

    def convert_test_suite(self):
        """
        遍历测试用例目录下的所有文件，类和方法，加载符合条件的测试方法（测试用例）;
        如果是文件，则直接加载
        :param: skip_prefix 目录和文件跳过的前缀
        :return: unittest.TestSuite 对象
        """
        if os.path.isdir(Settings.TEST_SUITE_PATH):
            for root, dir_, files in os.walk(Settings.TEST_SUITE_PATH):
                for file in files:
                    if not file.lower().startswith(self.skip_prefix) and file.split('.')[-1] == 'py':
                        abs_path = os.path.join(root, file)
                        import_list = self.get_import_list_by_path(abs_path)
                        if import_list is not None:
                            _import_list = import_list.split('.')
                            for sub_path in _import_list:
                                if sub_path.lower().startswith(self.skip_prefix):
                                    break
                            else:
                                package = import_module(import_list)
                                self.load_test_set(package)
        else:
            import_list = self.get_import_list_by_path(Settings.TEST_SUITE_PATH)
            if import_list is not None:
                for sub_path in import_list.split('.'):
                    if sub_path.lower().startswith(self.skip_prefix):
                        break
                else:
                    package = import_module(import_list)
                    self.load_test_set(package)
        return self.test_suite


class ExcelTestSet(TestSet):
    """从Excel中读取测试集"""
    def __init__(self, test_prefix, skip_prefix):
        super(ExcelTestSet, self).__init__(test_prefix, skip_prefix)
        try:
            self._excel = ExcelParser(Settings.TEST_EXCEL_PATH, 'TestCases')
        except FileNotFoundError:
            raise FileNotFoundError("Cannot found testcase file: {}".format(Settings.TEST_EXCEL_PATH))
        self.test_set = self.load_test_set()

    def load_test_set(self):
        """
        一次性读取所有测试用例
        :return: [dict(case), dict(case)...]
        """
        test_set = list()
        titles = [x for x in self._excel.get_row_values(1)]
        for row in range(2, self._excel.max_row + 1):
            test_case = dict()
            row_hidden = self._excel.worksheet.row_dimensions[row].hidden   # 获取该行是否被隐藏
            # 过滤被隐藏的行
            if row_hidden is False:
                values = [x for x in self._excel.get_row_values(row)]
                # 将行数据按照表头个数补齐None
                if len(values) < len(titles):
                    for i in range(len(titles) - len(values)):
                        values.append(None)
                for title, value in zip(titles, values):
                    if title == 'ParamsData':
                        params = self.convert_params(value)
                        test_case[title] = params
                    elif title == 'AssertResult':
                        asserts = self.convert_asserts(value)
                        test_case[title] = asserts
                    elif title == 'Tags':
                        tags = self.convert_tags(value)
                        test_case[title] = tags
                    else:
                        test_case[title] = value
                test_set.append(test_case)
        return test_set

    def convert_test_suite(self):
        """
        将excel表格中读取的测试用例生成 PUnittest 能识别的 TestSuite 对象，或者获取本地的所有测试文件
        :return: unittest.TestSuite 对象
        """
        if len(self.test_set) > 0:
            for test_case in self.test_set:
                _dir_name = test_case['TestDir'] if 'TestDir' in test_case else None
                _file_name = test_case['TestFile'] if 'TestFile' in test_case else None
                _cls_name = test_case['TestClass'] if 'TestClass' in test_case else None
                _case_name = test_case['TestCase'] if 'TestCase' in test_case else None
                try:
                    if not _file_name.lower().startswith(self.skip_prefix) \
                            or not _cls_name.lower().startswith(self.skip_prefix):
                        package = import_module('{0}.{1}'.format(_dir_name, _file_name))
                        cls = getattr(package, _cls_name)
                        for name, func in list(cls.__dict__.items()):
                            # 将装饰过的测试用例尾部序号移除
                            _name = re.sub(r'_#[\d]*', '', name)
                            # 将装饰过的测试用例首部还原
                            _name = re.sub(self.case_prefix + r'_\d{5}', self.case_prefix, _name)
                            if _name == _case_name:
                                case = cls(name)
                                self.test_suite.addTest(case)
                except Exception as e:
                    logger.error('Fail to load test case <{0}><{1}>: {2}'.format(_cls_name, _case_name, e))
        else:
            logger.error('Fail to load any test case, please check')
            raise RuntimeError('Fail to load any test case, please check')
        return self.test_suite


class ApiTestSet(TestSet):
    """从Http接口获取测试集"""
    def __init__(self, data, test_prefix, skip_prefix):
        super(ApiTestSet, self).__init__(test_prefix, skip_prefix)
        self.api_data = data
        self.test_set = self.load_test_set()

    def load_test_set(self):
        test_set = list()
        for item in self.api_data:
            test_case = dict()
            for k, v in item.items():
                if k == 'ParamsData':
                    params = self.convert_params(v)
                    test_case[k] = params
                elif k == 'AssertResult':
                    asserts = self.convert_asserts(v)
                    test_case[k] = asserts
                elif k == 'Tags':
                    tags = self.convert_tags(v)
                    test_case[k] = tags
                else:
                    test_case[k] = v
            test_set.append(test_case)
        return test_set

    def convert_test_suite(self):
        raise NotImplementedError("Not implemented yet...")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .settings import Settings as SETTINGS
from .utils.logger import logger
from .testrunner import TestRunner
from .testresult import TestResult
from .statistics import Statistics
from importlib import reload, import_module


def RELOAD_SETTINGS():
    SETTINGS.__SPECIFIED__ = True

    from . import settings
    settings.Settings = SETTINGS
    from .utils.logger import Logger
    global logger
    logger = Logger(
        logger_dir=SETTINGS.LOG_DIR,
        console_level=SETTINGS.LOG_CONSOLE_LEVEL,
        file_level=SETTINGS.LOG_FILE_LEVEL,
        report_level=SETTINGS.LOG_REPORT_LEVEL
    ).get_new_logger(
        console_switch=SETTINGS.LOG_CONSOLE_SWITCH,
        file_switch=SETTINGS.LOG_FILE_SWITCH,
        report_switch=SETTINGS.LOG_REPORT_SWITCH
    )

    from .testcase import TestCaseHandler
    TestCaseHandler.total_case_num = 0
    TestCaseHandler.actual_case_num = 0

    import punittest
    reload(punittest)
    from punittest import PUnittest

    # 通过递归的方式重载所有PUnittest的子类所在的文件（也就是测例文件），
    # 这样做的目的是可以在运行过程中想要修改测例的配置，比如tag
    def get_all_subclasses(cls):
        subclasses = cls.__subclasses__()
        for subclass in subclasses:
            module = import_module(subclass.__module__)
            reload(module)
            get_all_subclasses(subclass)

    get_all_subclasses(PUnittest)


__all__ = ['SETTINGS', 'RELOAD_SETTINGS', 'TestRunner', 'TestResult', 'logger', 'Statistics']


if SETTINGS.__SPECIFIED__ is True:
    from .punittest import PUnittest
    __all__.append('PUnittest')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os

from copy import deepcopy
from importlib import import_module
from .settings import Settings
from .testcase import CASE_SKIP_FLAG, CASE_TAG_FLAG
from .testset import CodeTestSet

KEYWORDS = ["testrunner", "testset", "testresult", "testcase", "testsuite"]


class Statistics:
    """由测试类继承，负责统计测试类中的数据"""
    def __init__(self):
        self.run_file = __file__
        self.suite_path = Settings.TEST_SUITE_PATH.lower()
        self.test_prefix = Settings.TEST_PREFIX.lower()
        self.skip_prefix = Settings.SKIP_PREFIX.lower()

    def _file_info(self, file_path):
        """根据file_path获取file_info"""
        dir_, name = os.path.split(file_path)[0], os.path.split(file_path)[1]
        if not name.startswith("_") and os.path.splitext(name)[-1] == ".py":
            file_info = {"file_name": name, "file_path": file_path}
            import_list = CodeTestSet.get_import_list_by_path(file_path)
            file_info["file_import_list"] = import_list
            if import_list is not None:
                _import_list = import_list.split('.')
                for sub_path in _import_list:
                    if sub_path.lower().startswith(self.skip_prefix):
                        file_info["file_skip"] = True
                        break
                else:
                    file_info["file_skip"] = False
            return file_info

    def _test_file(self):
        result = []
        if os.path.isdir(self.suite_path):
            for root, _, files in os.walk(self.suite_path):
                for file in files:
                    file_path = os.path.join(root, file)
                    file_info = self._file_info(file_path)
                    if file_info is not None:
                        result.append(file_info)
        else:
            file_info = self._file_info(self.suite_path)
            if file_info is not None:
                result.append(file_info)
        return result

    def _test_class(self, file_info):
        result = []
        module = import_module(file_info['file_import_list'])
        for cls_name in module.__dict__:
            if cls_name.lower().startswith(self.test_prefix) and cls_name.lower() not in KEYWORDS:
                cls_ = getattr(module, cls_name)
                file_info['class_name'] = cls_name
                file_info['class_skip'] = getattr(cls_, CASE_SKIP_FLAG, False)
                class_info = deepcopy(file_info)
                result.append(class_info)
        return result

    def _test_case(self, class_info):
        result = []
        module = import_module(class_info['file_import_list'])
        cls_ = getattr(module, class_info['class_name'])
        for func_name in cls_.__dict__:
            if func_name.lower().startswith(self.test_prefix):
                func = getattr(cls_, func_name)
                class_info['case_name'] = func_name
                class_info['case_skip'] = getattr(func, CASE_SKIP_FLAG, False)
                class_info['case_tag'] = getattr(func, CASE_TAG_FLAG, {'all'})
                case_info = deepcopy(class_info)
                result.append(case_info)
        return result

    @staticmethod
    def test_files():
        state = Statistics()
        files = state._test_file()
        for i, file_info in enumerate(files):
            file_skip = file_info.pop('file_skip')
            new_file = {'num': i + 1}
            new_file.update(file_info)
            new_file['if_skip'] = file_skip
            files[i] = new_file
        return files

    @staticmethod
    def test_classes():
        classes = []
        i = 1
        state = Statistics()
        for file_info in state._test_file():
            for class_info in state._test_class(file_info):
                file_skip = class_info.pop('file_skip')
                class_skip = class_info.pop('class_skip')
                new_class = {'num': i}
                new_class.update(class_info)
                new_class['if_skip'] = file_skip or class_skip
                classes.append(new_class)
                i += 1
        return classes

    @staticmethod
    def test_cases():
        cases = []
        i = 1
        state = Statistics()
        for file_info in state._test_file():
            for class_info in state._test_class(file_info):
                for case_info in state._test_case(class_info):
                    file_skip = case_info.pop('file_skip')
                    class_skip = case_info.pop('class_skip')
                    case_skip = case_info.pop('case_skip')
                    new_case = {'num': i}
                    new_case.update(case_info)
                    new_case['if_skip'] = file_skip or class_skip or case_skip
                    cases.append(new_case)
                    i += 1
        return cases



#! /usr/bin/env python
# -*- coding: UTF-8 -*-

import unittest
import re
import sys
import time

from .testset import CodeTestSet, ExcelTestSet, ApiTestSet
from .settings import Settings
from .utils.logger import logger
from functools import wraps

CASE_TAG_FLAG = "__case_tag__"
CASE_DATA_FLAG = "__case_data__"
CASE_ID_FLAG = "__case_id__"
CASE_INFO_FLAG = "__case_info__"
CASE_RUN_INDEX_FlAG = "__case_run_index_flag__"
CASE_SKIP_FLAG = "__unittest_skip__"
CASE_SKIP_REASON_FLAG = "__unittest_skip_why__"


def check_prefix(name, test_prefix=Settings.TEST_PREFIX):
    """
    检查名称的前缀是否符合test_prefix，忽略大小写
    :param name: 名称
    :param test_prefix: 前缀
    :return: bool
    """
    return name.lower().startswith(test_prefix)


def skip(reason):
    def wrap(func):
        return unittest.skip(reason)(func)
    return wrap


def skip_if(condition, reason):
    def wrap(func):
        return unittest.skipIf(condition, reason)(func)
    return wrap


def data(params, asserts):
    """注入测试数据，可以做为测试用例的数据驱动
    :param params: 测试数据组成的list
    :param asserts: 断言数据组成的list
    :return:
    """
    def wrap(func):
        param_data = {'params': params, 'asserts': asserts}
        setattr(func, CASE_DATA_FLAG, param_data)
        return func
    return wrap


def list_data(*datas):
    """注入测试数据，接收list作为参数，可以做为测试用例的数据驱动
    :param datas: 测试数据组成的list
    :return:
    """
    def wrap(func):
        param_data = {'datas': datas}
        setattr(func, CASE_DATA_FLAG, param_data)
        return func
    return wrap


def tag(*tag_type):
    """指定测试用例的标签，可以作为测试用例分组使用，用例默认会有Tag.al标签，支持同时设定多个标签，如：
    @tag(Tag.V1_0_0, Tag.SMOKE)
    def test_func(self):
        pass

    :param tag_type:标签类型，在tag.py里边自定义
    :return:
    """
    def wrap(func):
        _tag_type = [_tag.lower() for _tag in tag_type]
        if not hasattr(func, CASE_TAG_FLAG):
            tags = {'all'}
            tags.update(_tag_type)
            setattr(func, CASE_TAG_FLAG, tags)
        else:
            getattr(func, CASE_TAG_FLAG).update(_tag_type)
        return func
    return wrap


def _assertion(cap_func=None, cap_dir=None, arg_length=None, **cap_kwargs):
    """用来装饰PUnittest类中所有的AssertXxx方法，提供日志，和失败截图，
        并且当出现断言异常不会立即终止用例，而是等用例结束后抛出所有的断言异常"""
    def wrapper(func):
        @wraps(func)
        def wrap(*args, **kwargs):
            assert_args = []
            for assert_arg in args[1:]:
                if isinstance(assert_arg, bytes):
                    assert_args.append("bytes...")
                elif isinstance(assert_arg, bytearray):
                    assert_args.append("bytearray...")
                elif isinstance(arg_length, int) and 0 < arg_length < len(str(assert_arg)):
                    assert_args.append(str(assert_arg)[:arg_length] + '...')
                else:
                    assert_args.append(assert_arg)
            assert_args = tuple(assert_args)
            try:
                logger.info('[Assert]: {0}{1}'.format(func.__name__, format(assert_args)))
                return func(*args, **kwargs)
            except AssertionError as e:
                args[0].Exc_Stack.append(e)
                if cap_func is not None and cap_dir is not None:
                    cap_name = "{}_{}.png".format(
                        time.strftime("%Y-%m-%d_%H-%M-%S"), getattr(args[0], '_testMethodName')
                    )
                    cap_func(cap_dir, cap_name, **cap_kwargs)
        return wrap
    return wrapper


def _handler(rerun=Settings.CASE_FAIL_RERUN, cap_func=None, cap_dir=None, **cap_kwargs):
    """处理测试用例，为用例添加格式化的日志。并提供失败重跑机制。"""
    def wrapper(func):
        @wraps(func)
        def wrap(*args, **kwargs):
            _testcase_name = getattr(args[0], '_testMethodName')
            _testcase_doc = getattr(args[0], '_testMethodDoc')
            _testclass_name = args[0].__class__.__name__
            _testclass_doc = args[0].__class__.__doc__
            rerun_time = rerun
            while rerun_time > 0:
                try:
                    logger.info('[TestCase]: {0}.{1} {2}|{3}'.format(
                        _testclass_name, _testcase_name, _testclass_doc, _testcase_doc))
                    case_tags = ', '.join(sorted(list(getattr(func, CASE_TAG_FLAG, {'all'}))))
                    logger.info('[CaseTags]: {0}'.format(case_tags))
                    logger.info('[TestProgress]: {0}/{1}\t[RerunTime]: NO.{2}'.format(
                        getattr(func, CASE_RUN_INDEX_FlAG), TestCaseHandler.actual_case_num, rerun - rerun_time + 1))
                    args[0].clear_exc_stack()
                    result = func(*args, **kwargs)
                    # 用例执行完毕抛出所有可能存在的AssertionError异常
                    args[0].raise_exc()
                    logger.info('[TestResult]: Pass!')
                    return result
                except AssertionError as e:
                    rerun_time -= 1
                    if rerun_time <= 0:
                        logger.error('[TestResult]: Fail!')
                        raise e
                except Exception as e:
                    rerun_time -= 1
                    if rerun_time <= 0:
                        _, _, exc_traceback = sys.exc_info()
                        logger.error('[TestResult]: Error!')
                        if cap_func is not None and cap_dir is not None:
                            cap_name = "{}_{}.png".format(
                                time.strftime("%Y-%m-%d_%H-%M-%S"), getattr(args[0], '_testMethodName')
                            )
                            cap_func(cap_dir, cap_name, **cap_kwargs)
                        raise e
        return wrap
    return wrapper


def _feed_data(*args, **kwargs):
    def wrap(func):
        @wraps(func)
        def _wrap(self):
            return func(self, *args, **kwargs)
        return _wrap
    return wrap


class TestCaseHandler:
    actual_case_num = 0
    total_case_num = 0

    @classmethod
    def create_case_id(cls):
        cls.total_case_num += 1
        return cls.total_case_num

    @classmethod
    def create_actual_run_index(cls):
        cls.actual_case_num += 1
        return cls.actual_case_num

    @staticmethod
    def modify_func_name(func, index):
        """修改函数名字，实现排序，并且根据DDT数据的编号在函数名后添加编号 eg test_fight >> test_00001_fight_#1
        :param func:
        :param index:
        :return:
        """
        case_id = TestCaseHandler.create_case_id()
        setattr(func, CASE_ID_FLAG, case_id)
        func_name = func.__name__.replace(
            Settings.TEST_PREFIX, Settings.TEST_PREFIX + "_{:05d}".format(case_id)
        )
        if index is not None:
            func_name += "_#{0}".format(index)
        return func_name

    @staticmethod
    def _single_case_handler(func, func_index):
        """处理单条测试用例，注入数据，修改名称等，返回装饰器"""
        if not hasattr(func, CASE_SKIP_FLAG):
            setattr(func, CASE_RUN_INDEX_FlAG, TestCaseHandler.create_actual_run_index())
        func_name = TestCaseHandler.modify_func_name(func, func_index)
        func_handler = _handler(
            cap_func=Settings.CAP_FUNC,
            cap_dir=Settings.CAP_DIR,
            **Settings.CAP_KWARGS
        )
        return func_name, func_handler

    @staticmethod
    def create_case_with_case_data(func):
        """根据原有测试函数创建新的带参数的测试函数"""
        result = dict()
        test_data = getattr(func, CASE_DATA_FLAG)
        params = test_data.get('params', [])
        asserts = test_data.get('asserts', [])
        datas = test_data.get('datas', [])
        if params and asserts:
            for index, (param, assert_) in enumerate(zip(params, asserts), 1):
                func_name, func_handler = TestCaseHandler._single_case_handler(func, index)
                result[func_name] = func_handler(_feed_data(params=param, asserts=assert_)(func))
        elif datas:
            for index, data_ in enumerate(datas, 1):
                func_name, func_handler = TestCaseHandler._single_case_handler(func, index)
                result[func_name] = func_handler(_feed_data(datas=data_)(func))
        return result

    @staticmethod
    def create_case_without_case_data(func):
        """根据原有测试函数创建新的不带参数的测试函数"""
        result = dict()
        func_name, func_handler = TestCaseHandler._single_case_handler(func, None)
        result[func_name] = func_handler(func)
        return result

    @staticmethod
    def filter_cases_by_tags(tags, func):
        """根据func的tags进行过滤"""
        target_tags = [x.lower() for x in tags]
        name = func.__name__
        if check_prefix(name):
            have_tags = getattr(func, CASE_TAG_FLAG, {'all'})
            for target_tag in target_tags:
                for have_tag in have_tags:
                    pattern = "^" + target_tag.replace("*", ".*?") + "$"
                    result = re.match(pattern, have_tag)
                    if result is not None:
                        return func
            else:
                msg = "不包含需要执行的Tag: {}".format(', '.join(tags))
                filtered_func = skip(msg)(func)
                return filtered_func
        return func


def get_test_set(api_data=None):
    """获取测试用例集"""
    if Settings.TEST_SET_FORM == "CODE":
        test_set = CodeTestSet(Settings.TEST_PREFIX, Settings.SKIP_PREFIX).test_set
    elif Settings.TEST_SET_FORM == "EXCEL":
        test_set = ExcelTestSet(Settings.TEST_PREFIX, Settings.SKIP_PREFIX).test_set
    elif Settings.TEST_SET_FORM == "API":
        test_set = ApiTestSet(api_data, Settings.TEST_PREFIX, Settings.SKIP_PREFIX).test_set
    else:
        test_set = CodeTestSet(Settings.TEST_PREFIX, Settings.SKIP_PREFIX).test_set
    return test_set


class MataClassHandler:

    @staticmethod
    def classify_funcs(funcs_dict):
        """
        将类中的函数进行归类
        :param funcs_dict: 函数字典
        :return:
        """
        funcs = dict()
        cases = dict()
        asserts = dict()
        for attr in funcs_dict:
            if check_prefix(attr):
                cases[attr] = funcs_dict[attr]
            elif attr.startswith("assert"):
                if attr not in ["assertMultiLineEqual", "assertIsInstance"]:
                    asserts[attr] = funcs_dict[attr]
                else:
                    funcs[attr] = funcs_dict[attr]
            else:
                funcs[attr] = funcs_dict[attr]
        return funcs, cases, asserts

    @staticmethod
    def bind_case_skip_tag(cls_name, cases, skip_prefix):
        """
        检查每条测试用例的类和模块是否包含skip_prefix，
        如果包含则将该case标记为skip
        :param cls_name: 类名
        :param cases: 测试用例字典
        :param skip_prefix: 测试模块和测试类中跳过的前缀，来源于 Settings
        :return:
        """
        skip_prefix = skip_prefix.lower()

        if len(cases) > 0:
            module_name = list(cases.values())[0].__module__
        else:
            module_name = ""

        if cls_name.lower().startswith(skip_prefix):
            for k, v in cases.items():
                cases[k] = skip("{} 模块跳过".format(module_name))(v)
        elif module_name.lower().startswith(skip_prefix):
            for k, v in cases.items():
                cases[k] = skip("{} 测试类跳过".format(cls_name))(v)
        return cases

    @staticmethod
    def handle_asserts(funcs, asserts):
        """
        装饰测试类中的每个断言函数，将其放入funcs
        :param funcs: 普通函数字典
        :param asserts: 断言函数字典
        :return:
        """
        for name, assert_ in asserts.items():
            # 将装饰过的assert方法加入类
            funcs.update(
                {name: _assertion(
                    cap_func=Settings.CAP_FUNC,
                    cap_dir=Settings.CAP_DIR,
                    arg_length=Settings.ASSERT_ARG_LENGTH,
                    **Settings.CAP_KWARGS)(assert_)}
            )
        return funcs

    @staticmethod
    def handle_case_by_code(funcs, cases, tags):
        """
        直接处理代码中测试类中的每个函数，将每个测试函数处理为测试用例
        :param funcs: 测试类中的普通函数，和装饰后的assert函数
        :param cases: 测试类中以prefix开头的测试函数（测试用例）
        :param tags: 需要运行的测试用例的tag
        :return: funcs更新装饰后的测试函数
        """
        for test_func in cases.values():
            # 根据tag进行过滤
            test_func = TestCaseHandler.filter_cases_by_tags(tags, test_func)
            # 注入用例信息
            case_info = "{}.{}".format(test_func.__module__, test_func.__name__)
            setattr(test_func, CASE_INFO_FLAG, case_info)
            # 注入测试数据
            if hasattr(test_func, CASE_DATA_FLAG):
                funcs.update(TestCaseHandler.create_case_with_case_data(test_func))
            else:
                funcs.update(TestCaseHandler.create_case_without_case_data(test_func))
        return funcs

    @staticmethod
    def handle_case_by_test_set(funcs, cases, tags, cls_name, test_set):
        """
        根据外在测试集处理测试类中的每个函数，将每个测试函数处理为测试用例
        :param funcs: 测试类中的普通函数，和装饰后的assert函数
        :param cases: 测试类中以prefix开头的测试函数（测试用例）
        :param tags: 需要运行的测试用例的tag
        :param cls_name: 类名称
        :param test_set: 测试集
        :return: funcs更新装饰后的测试函数
        """
        for test_case in test_set:
            _cls_name = test_case['TestClass'] if 'TestClass' in test_case else None
            _case_name = test_case['TestCase'] if 'TestCase' in test_case else None
            _case_desc = test_case['CaseDescription'] if 'CaseDescription' in test_case else None
            _tags = test_case['Tags'] if 'Tags' in test_case else None
            _skip = test_case['Skip'] if 'Skip' in test_case else None
            _params = test_case['ParamsData'] if 'ParamsData' in test_case else None
            _asserts = test_case['AssertResult'] if 'AssertResult' in test_case else None
            if cls_name == _cls_name:
                for test_func in cases.values():
                    if test_func.__name__ == _case_name:
                        # 为测试函数绑定tag装饰器
                        test_func = tag(*_tags)(test_func)
                        # 根据tag进行过滤
                        test_func = TestCaseHandler.filter_cases_by_tags(tags, test_func)
                        # 为测试函数绑定skip装饰器，只认定excel中的skip标识，忽略编码中的skip标识
                        if _skip is not None:
                            test_func = skip(_skip)(test_func)
                        # 修改测试函数的文档注释
                        if _case_desc:
                            setattr(test_func, '__doc__', _case_desc)
                        # 为测试函数绑定data装饰器
                        if _asserts is not None and _params is not None:
                            test_func = data(_params, _asserts)(test_func)
                            funcs.update(TestCaseHandler.create_case_with_case_data(test_func))
                        else:
                            funcs.update(TestCaseHandler.create_case_without_case_data(test_func))
        return funcs

    @staticmethod
    def filter_skipped_class(cls, funcs):
        """
        如果一个testclass中的所有testfunc都跳过不执行，则该testclass也被标记为跳过，
        setupClass和teardownClass也不执行
        :param cls: 测试类
        :param funcs: 包括所有需要执行的函数
        :return: 更新后的测试类
        """
        if_skip = True
        for func_name, func_obj in funcs.items():
            if check_prefix(func_name):
                if not hasattr(func_obj, CASE_SKIP_FLAG):
                    if_skip = False
                    break
        if check_prefix(cls.__name__):
            setattr(cls, CASE_SKIP_FLAG, if_skip)
        return cls


class Meta(type):
    def __new__(mcs, cls_name, supers, cls_dict):
        test_set = get_test_set()
        tags = Settings.RUN_TAGS
        skip_prefix = Settings.SKIP_PREFIX
        funcs, cases, asserts = MataClassHandler.classify_funcs(cls_dict)
        # 根据测试模块和测试类的名称是否包含跳过前缀，为测试用例绑定skip装饰器
        cases = MataClassHandler.bind_case_skip_tag(cls_name, cases, skip_prefix)
        # 处理断言函数后加入funcs
        funcs = MataClassHandler.handle_asserts(funcs, asserts)
        # 处理测试用例后加入funcs
        if test_set is None:
            funcs = MataClassHandler.handle_case_by_code(funcs, cases, tags)
        else:
            funcs = MataClassHandler.handle_case_by_test_set(funcs, cases, tags, cls_name, test_set)
        cls = super(Meta, mcs).__new__(mcs, cls_name, supers, funcs)

        cls = MataClassHandler.filter_skipped_class(cls, funcs)
        return cls
